(* Hilbert og Sierpinski kurvetegning *)

structure Curves :> Curves = 
struct


(*************************************)
(* Kurve-funktioner                  *)
(*************************************)


(* Grundlggende datatyper - point reprsenterer et punkt i planen, curve en kurve af rette linjer *)
type point = (real*real);
type curve = point list 


(* Point: point -> curve
 * Danner en kurve af et enkelt punkt *)
fun Point( p: point ) = [ p ];


(* Line: point * point -> curve
 * Danner kurven bestende af den rette linje melle p og q *)
fun Line( p: point, q: point ) = [ p, q ];


(* Join: curve * curve -> curve
 * Danner kurven sammensat af de to kurver cl og dl med en ret linje *)
fun Join( cl: curve, dl: curve ) = cl @ dl;


(* Trans: curve * point -> curve
 * Flytter kurven med den forskydelse angivet i vektoren *)
fun Trans( c: curve, (vx,vy): point ) = 
    let val f = fn (x,y): point => (x+vx,y+vy) in
	map f ( c )
   end


(* Rot: curve * real -> curve
 * Roterer kurven c omkring origo det angivne antal grader *)
fun Rot( c: curve, deg: real ) =
    let 
	val angle = deg*Math.pi/180.0;
	val f = fn (x,y):point => (x*Math.cos(angle)-y*Math.sin(angle), x*Math.sin(angle)+y*Math.cos(angle))
    in
	map f (c)
    end


(* Scale: curve * real -> curve
 * Skalerer kurven udfra origo *)
fun Scale( c: curve, fact: real ) =
    let val f = fn (x,y):point => (x*fact, y*fact) in
	map f (c)
    end


(* ReflectX: curve -> curve
 * Spejler kurven i X-aksen *)
fun ReflectX( c: curve ) =
    let val f = fn (x,y):point => (x,~y) in
	map f (c)
    end


(* ReflectY: curve -> curve
 * Spejler kurven i Y-aksen *)
fun ReflectY( c: curve ) = 
    let val f = fn (x,y):point => (~x,y) in
	map f (c)
    end


(* SizeX: curve -> real
 * Bestemmer horisontal strrelse af kurven *)
fun SizeX( [] ) = 0.0
 |  SizeX( (x1,_)::cl : curve ) = 
    let 
	val maxX = fn ((x,_):point, m:real) => if x>m then x else m;
	val minX = fn ((x,_):point, m:real) => if x<m then x else m 
    in
        foldl maxX x1 (cl) - foldl minX x1 (cl) 
    end


(* SizeY: curve -> real
 * Bestemmer vertikal strrelse af kurven *)
fun SizeY( [] ) = 0.0
 |  SizeY( (_,y1)::cl : curve ) = 
    let 
	val maxY = fn ((_,y):point, m:real) => if y>m then y else m;
	val minY = fn ((_,y):point, m:real) => if y<m then y else m 
    in
        foldl maxY y1 (cl) - foldl minY y1 (cl) 
    end




(*************************************)
(* Transformationer                  *)
(*************************************)


(* Hilbert udfrer de 4 transformationer, der danner Hilbert-kurven. De 4 delkurver sammensttes til den endelige kurve.
 *)
fun Hilbert( [] ) = []
 |  Hilbert( st::h ) = 
    let
	val c = Trans( st::h, st);
        val aa = Rot( ReflectY(c), ~90.0 );
	val ab = Trans(c, (0.0,SizeY(c)+1.0) );
        val bb = Trans(c, (SizeX(c)+1.0, SizeY(c)+1.0) );
        val ba = Trans( Rot( ReflectY(c), 90.0 ), (2.0*SizeX(c)+1.0,SizeY(c)))
    in
	Join( Join(Join(aa,ab),bb),ba)
    end


(* Sierpinski funktionen gennemfrer de 4 transformationer af kurven, der udgr Sierpinski-kurven.
 * De 4 kurver, der dannes sammensttes til den endelige Hilbert-kurve.
 *)
fun Sierpinski( [] ) = []
 |  Sierpinski( st::s ) = 
    let 
	val sn = Trans( st::s, st);
        val sn1 = sn;
	val sn2 = Trans( Rot(sn, 90.0), (SizeY(sn), SizeY(sn)+1.0));
        val sn3 = Trans( Rot(sn, ~90.0), (1.0, 2.0*(SizeY(sn)+1.0)));
        val sn4 = Trans( sn, (SizeX(sn)+2.0, SizeY(sn)+2.0))
    in
	Join(Join(Join( sn1, sn2), sn3), sn4)
    end


(* Peano funktionen gennemfrer de 9 transformationer af den angivne kurve (reduceres til 4 ved genbrug), der danner Peano-kurven.
 
fun Peano( [] ) = []
 |  Peano( ps::p ) = 
    let
	val pn = Trans( ps::p, ps);
	val s1 = pn;
	val s2 = Trans(ReflectX(pn), (SizeX(pn)+1.0,SizeY(pn)));
	val s3 = Trans(pn, ((SizeX(pn)+1.0)*2.0,0.0));
	val p1 = Join(Join(s1,s2),s3);
	val p2 = Trans( ReflectY(p1), (3.0*SizeX(pn)+2.0,SizeY(pn)+1.0));
	val p3 = Trans( p1, (0.0, 2.0*SizeY(pn)+2.0))
    in
	Join( Join(p1, p2), p3)
    end

*)


(* SmartRec tegner den angivne kurve ved at gennemfre en smart rekursion med anvendelse af den angivne transformation p kurven 1
 * niveau under. Dette gr det muligt at tegne de rekursive kurver nemt. F.eks. vil dannelsen af Hilbert p niveau 5 kunne skrives:
 * SmartRec Hilbert (5)
 *)
fun SmartRec Transform (n) = 
  if n<=0 then Point(0.0,0.0)
  else Transform( SmartRec Transform (n-1) );



(*************************************)
(* PostScript output funktioner      *)
(*************************************)



(* realToString: real -> string
 * Returnerer korrekt formateret streng med vrdi *)
fun realToString(x) = 
    if (abs(x*1e8)<1.0) then "0.0" else
	if (x<0.0) then "-" ^ Real.toString(~x) 
	else Real.toString(x);


(* Draw: curve -> string
 * Beregner en streng af PostScript kommandoer, der kan skrives til fil eller printes ud p PS-printer
 * Denne funktion br ikke bruges, da den er YDERST hukommelseskrvende! 
 * Brug curveToPostscript
 *)
fun Draw( [] ) = "" 
 |  Draw( _::[] ) = ""
 |  Draw( (x1,y1)::(x2,y2)::cl : curve ) =
      realToString(x1) ^ " " ^ realToString(y1) ^ " moveto\n" ^ realToString(x2) ^ " " ^ realToString(y2) ^ " lineto\nstroke\n" ^ Draw((x2,y2)::cl);


(* toPostscript: string * string -> unit
 * Skriver s (bestende af postscript-kommandoer) ud til filen angivet i filename
 * Bruges i forbindelse med Draw 
 * OUTDATED - brug i stedet curveToPostscript
 *)

fun toPostscript( s: string, filename: string ) =
    let 
	val out = TextIO.openOut(filename);
	val data = "%! PS-Adobe-2.0\n%%BoundingBox: 100.0 100.0 200.0 186.0\n1 1 scale\nnewpath\n\n" ^ s ^ "\nshowpage\n\n"
    in
	TextIO.output(out,data);
	TextIO.closeOut(out)
    end


(* curveToPostscript: curve * string -> unit
 * Skriver den angivne kurve ud i filen angivet. Br bruges i stedet for Draw-toPostscript kombinationen,
 * der bruger p... meget hukommelse *)
local
  fun Draw2( out, [] ) = TextIO.output(out, "")
   |  Draw2( out, _::[] ) = TextIO.output(out, "")
   |  Draw2( out, (x1,y1)::(x2,y2)::cl : curve ) =
      let
        val lineseg = realToString(x1) ^ " " ^ realToString(y1) ^ " moveto\n" ^ realToString(x2) ^ " " ^ realToString(y2) ^ " lineto\nstroke\n"
      in
        TextIO.output(out, lineseg);
        Draw2( out, (x2,y2)::cl)
      end;
in fun curveToPostscript( c: curve, filename: string ) =
    let 
	val out = TextIO.openOut(filename);
	val head = "%! PS-Adobe-2.0\n%%BoundingBox: 0.0 0.0 " ^ realToString(SizeX(c)) ^ " " ^ realToString(SizeY(c)) ^ "\n1 1 scale\nnewpath\n\n";
	val foot = "\nshowpage\n\n"
    in
       TextIO.output(out, head);
       Draw2( out, c );
       TextIO.output(out, foot);
       TextIO.closeOut(out)
    end
end;





(* DrawOneCurve tegner den angivne transformation p det angivne niveau skaleret korrekt i en postscript fil.
 *)
fun DrawOneCurve Transform (n: int, fname: string) = 
    let
	val c = SmartRec Transform (n);
	val scX = 580.0/(SizeX(c)+1.0);
	val scY = 730.0/(SizeY(c)+1.0);
	val sc = Trans(Scale(c, Real.min(scX,scY)), (20.0,20.0));
    in
	curveToPostscript(sc, fname)
    end;



end;



(* Kommandolinje:

compile "Curves.sig";
compile "Curves.sml";
load "Curves";
open Curves;

*)


